// C/C++ Headers

#include <iostream.h>
#include <string.h>

#ifdef _WIN32
#   include <float.h>
#   define ISNAN(x) _isnan(x)
#else
#   include <math.h>
#   define ISNAN(x) isnan(x)
#endif

// Local Headers

#include "DoubleVariables.h"
#include "tf.h"

/*****************************************************************************
 * Initialize static members of class 'DoubleVariables'
 *****************************************************************************/

const double DoubleVariables::ZERO = 0.0;

/*****************************************************************************
 * Declaration of private struct 'DoubleEntry'
 *****************************************************************************/

struct DoubleEntry
{
   DoubleEntry(const char *name);
   ~DoubleEntry();

   const char *mName;
   double mValue;

   ostream &print(ostream &out) const;
};

ostream &operator<<(ostream &out, const DoubleEntry &entry);

/*****************************************************************************
 * Definition of private struct 'DoubleEntry'
 *****************************************************************************/

DoubleEntry::DoubleEntry(const char *name) :
   mName(strcpy(new char[strlen(name) + 1], name)),
   mValue(DoubleVariables::ZERO)
{
}

DoubleEntry::~DoubleEntry()
{
   delete [] (char *)mName;
}

ostream &DoubleEntry::print(ostream &out) const
{
   out << mName << " = [" << mValue << "]";
   return(out);
}

ostream &operator<<(ostream &out, const DoubleEntry &entry)
{
   return(entry.print(out));
}

/*****************************************************************************
 * Definition of class 'DoubleVariables'
 *****************************************************************************/

DoubleVariables::DoubleVariables() :
   mCount(0),
   mEntries(0)
{
}

DoubleVariables::~DoubleVariables()
{
   for(int i = 0; i < mCount; i ++)
      delete mEntries[i];

   delete [] mEntries;
}

int DoubleVariables::assign(const char *name)
{
   int handle = lookup(name);

   if(handle == UNDEFINED)
   {
      DoubleEntry **newEntries = new DoubleEntry *[mCount + 1];

      for(int i = 0; i < mCount; i ++)
         newEntries[i] = mEntries[i];

      newEntries[mCount] = new DoubleEntry(name);
      delete [] mEntries;
      mEntries           = newEntries;

      handle = mCount++;
   }

   return(handle);
}

int DoubleVariables::lookup(const char *name) const
{
   for(int i = 0; i < mCount; i ++)
      if(strcmp(name, mEntries[i]->mName) == 0)
         return(i);

   return(UNDEFINED);
}

const char *DoubleVariables::getName(int handle) const
{
   if(!checkHandle(handle))
      return("null");

   return(mEntries[handle]->mName);
}

void DoubleVariables::initialize()
{
   for(int i = 0; i < mCount; i ++)
      set(i, ZERO);
}

void DoubleVariables::set(int handle, double value)
{
   if(checkHandle(handle))
   {
      // Apparently on win32 the value 'nan' is basically treated
      // as being zero (0.0).  Since any calculation involving
      // nan is basically meaningless anyway we do the check here
      // using 'isnan' to make the calculation come out the same
      // on linux as it does on win32.

      mEntries[handle]->mValue = (ISNAN(value) ? ZERO : value);
   }
}

double DoubleVariables::get(int handle) const
{
   return(checkHandle(handle) ? mEntries[handle]->mValue : ZERO);
}

ostream &DoubleVariables::print(ostream &out) const
{
   for(int i = 0; i < mCount; i ++)
   {
      if(i > 0)
         out << "\n";

      out << "entry[" << i << "]" << " => ";
      mEntries[i]->print(out);
   }

   return(out);
}

int DoubleVariables::main(int, char **)
{
   DoubleVariables variables;

   int xHandle = variables.assign("x");
   cout << "xHandle is: [" << xHandle << "]" << endl;

   int yHandle = variables.assign("y");
   cout << "yHandle is: [" << yHandle << "]" << endl;

   int zHandle = variables.assign("z");
   cout << "zHandle is: [" << zHandle << "]" << endl;

   cout << "name for xHandle is: [" <<
      variables.getName(xHandle) << "]" << endl;

   cout << "name for yHandle is: [" <<
      variables.getName(yHandle) << "]" << endl;

   cout << "name for zHandle is: [" <<
      variables.getName(zHandle) << "]" << endl;

   cout << "lookup(z) => [" << variables.lookup("z") << "]" << endl;

   cout << "lookup(a) => [" << variables.lookup("a") << "]" << endl;

   variables.set(xHandle, 1.0);
   variables.set(yHandle, 2.0);
   variables.set(zHandle, 3.0);

   cout << variables << endl;

   variables.initialize();

   cout << variables << endl;

   return(0);
}

int DoubleVariables::checkHandle(int handle) const
{
   if(handle >= 0 && handle < mCount)
      return(TRUE);

   cout <<
      "DoubleVariables::checkHandle(" << handle << ") - invalid" << endl;

   return(FALSE);
}

ostream &operator<<(ostream &out, const DoubleVariables &dv)
{
   return(dv.print(out));
}

#ifdef __MAIN__

int main(int argc, char **argv)
{
   return(DoubleVariables::main(argc, argv));

}  /* main */

#endif
